"""
This module defines the device classes to be used to connect to and control the Hexapod PUNA from
Symétrie.

"""
import time
from datetime import datetime
from datetime import timedelta

from egse.device import DeviceConnectionState
from egse.device import DeviceInterface
from egse.hexapod import HexapodError
from egse.hexapod.symetrie import logger
from egse.hexapod.symetrie import pmac
from egse.hexapod.symetrie.alpha import AlphaControllerInterface
from egse.hexapod.symetrie.hexapod import HexapodSimulator
from egse.hexapod.symetrie.pmac import PMACError
from egse.hexapod.symetrie.pmac import PmacEthernetInterface
from egse.hexapod.symetrie.pmac import decode_Q29
from egse.proxy import Proxy
from egse.registry.client import RegistryClient
from egse.settings import Settings
from egse.zmq_ser import connect_address

DEVICE_SETTINGS = Settings.load(filename="puna.yaml")

NUM_OF_DECIMALS = 6  # used for rounding numbers before sending to PMAC


class PunaInterface(AlphaControllerInterface, DeviceInterface):
    """
    Interface definition for the PunaController, the PunaProxy and the PunaSimulator.
    """


class PunaController(PunaInterface):
    """
    The PunaController class allows controlling a Symétrie PUNA Hexapod through an Ethernet
    interface that is connecting a Symétrie Controller.

    The Symétrie Controller can be either in simulation mode or have a real Hexapod
    connected.

    **Synopsis**

        from egse.hexapod.symetrie.puna import PunaController
        hexapod = PunaController(hostname="10.33.178.145", port=1025)
        try:
            hexapod.connect()

            # do some useful things here with the hexapod

        except HexapodError as exc:
            print(exc)
        finally:
            hexapod.disconnect()

    The constructor also sets the connection parameters and tries to connect
    to the controller. Make sure that you explicitly use the hexapod.disconnect()
    command when no connection is needed anymore.

    The controller can also be used as a context manager, in which case the `connect()`
    and `disconnect()` methods should not be called:

        with PunaController() as puna:
            puna.info()

    """

    SPEC_POS_MAINTENANCE = 0
    """Hexapod specific position, for maintenance or Jog only."""
    SPEC_POS_ZERO = 1
    """Hexapod zero position."""
    SPEC_POS_RETRACTED = 2
    """Hexapod retracted position."""

    def __init__(self, hostname=None, port=None):
        """
        Opens a TCP/IP socket connection with the Hexapod PUNA Hardware Controller.

        Args:
            hostname (str): the IP address or fully qualified hostname of the Hexapod hardware
            controller.
                The default is defined in the ``settings.yaml`` configuration file.

            port (int): the IP port number to connect to, by default set in the ``settings.yaml``
            configuration file.

        Raises:
            HexapodError: when the connection could not be established for some reason.
        """

        super().__init__()

        if hostname is None or port is None:
            raise ValueError(f"Please provide both hostname and port for the PunaController, {hostname=}, {port=}")

        logger.debug(f"Initializing PunaController with hostname={hostname} on port={port}")

        try:
            self.pmac = PmacEthernetInterface()
            self.pmac.setConnectionParameters(hostname, port)
        except PMACError as exc:
            logger.warning(
                f"HexapodError: Couldn't establish connection with the Hexapod PUNA Hardware "
                f"Controller: ({exc})"
            )

    def is_simulator(self):
        return False

    def is_connected(self):
        return self.pmac.isConnected()

    def connect(self):
        try:
            self.pmac.connect()
        except PMACError as exc:
            logger.warning(f"PMACError caught: Couldn't establish connection ({exc})")
            raise ConnectionError("Couldn't establish a connection with the Hexapod.") from exc

        self.notify_observers(DeviceConnectionState.DEVICE_CONNECTED)

    def disconnect(self):
        try:
            self.pmac.disconnect()
        except PMACError as exc:
            raise ConnectionError("Couldn't disconnect from Hexapod.") from exc

        self.notify_observers(DeviceConnectionState.DEVICE_NOT_CONNECTED)

    def reconnect(self):
        if self.is_connected():
            self.disconnect()
        self.connect()

    def is_in_position(self):
        try:
            out = self.pmac.getQVars(36, [0], int)
        except PMACError as exc:
            raise HexapodError(
                "Couldn't retrieve information from Hexapod PUNA Hardware Controller.") from exc
        return bool(out[0] & 0x04)

    def info(self):
        try:
            msg = "Info about the Hexapod PUNA:\n"
            msg += f"model   = {self.pmac.getPmacModel()}\n"
            msg += f"CID     = {self.pmac.getCID()}\n"
            msg += f"version = {self.pmac.getVersion()}\n"
            msg += f"cpu     = {self.pmac.getCPU()}\n"
            msg += f"type    = {self.pmac.getType()}\n"
            msg += f"vendorID= {self.pmac.getVID()}\n"
            msg += f"date    = {self.pmac.getDate()}\n"
            msg += f"time    = {self.pmac.getTime()}\n"
            msg += f"today   = {self.pmac.getToday()}\n"
        except PMACError as exc:
            raise HexapodError(
                "Couldn't retrieve information from Hexapod PUNA Hardware Controller."
            ) from exc

        return msg

    def stop(self):
        try:
            rc = self.pmac.sendCommand(pmac.CMD_STOP)
        except PMACError as exc:
            raise HexapodError("Couldn't complete the STOP command.") from exc

        return rc

    def homing(self):
        try:
            rc = self.pmac.sendCommand(pmac.CMD_HOMING)
        except PMACError as exc:
            raise HexapodError("Couldn't complete the HOMING command.") from exc

        logger.info("Homing: Command was successful")

        return rc

    def is_homing_done(self):
        try:
            rc = self.pmac.getQVars(26, [0], int)[0]
        except PMACError as pmac_exc:
            logger.error(f"PMAC Exception: {pmac_exc}", exc_info=True)
            return False

        msg = {  # noqa: F841
            0: "Homing status is undefined.",
            1: "Homing is in progress",
            2: "Homing is done",
            3: "An error occurred during the Homing process.",
        }

        if rc == 2:
            return True

        return False

    def set_virtual_homing(self, tx, ty, tz, rx, ry, rz):
        try:
            rc = self.pmac.sendCommand(
                pmac.CMD_VIRTUAL_HOMING, tx=tx, ty=ty, tz=tz, rx=rx, ry=ry, rz=rz
            )
        except PMACError as exc:
            raise HexapodError("Couldn't execute the virtual homing command.") from exc

        logger.warning(
            f"Virtual Homing successfully set to {tx:.6f}, {ty:.6f}, {tz:.6f}, {rx:.6f}, "
            f"{ry:.6f}, {rz:.6f}."
        )

        return rc

    def activate_control_loop(self):
        try:
            rc = self.pmac.sendCommand(pmac.CMD_CONTROLON)
        except PMACError as exc:
            raise HexapodError("Couldn't activate the control loop.") from exc

        msg = {  # noqa: F841
            0: "Command was successful",
            -1: "Command was ignored",
            -2: "Control of the servo motors has failed",
        }

        return rc

    def deactivate_control_loop(self):
        try:
            rc = self.pmac.sendCommand(pmac.CMD_CONTROLOFF)
        except PMACError as exc:
            raise HexapodError("Couldn't de-activate the control loop.") from exc

        return rc

    def __move(self, cm, tx, ty, tz, rx, ry, rz):
        """
        Ask the controller to perform the movement defined by the arguments.

        For all control modes cm, the rotation centre coincides with the Object
        Coordinates System origin and the movements are controlled with translation
        components at first (Tx, Ty, tZ) and then the rotation components (Rx, Ry, Rz).

        Control mode cm:
            * 0 = absolute control, object coordinate system position and orientation
                    expressed in the invariant user coordinate system
            * 1 = object relative, motion expressed in the Object Coordinate System
            * 2 = user relative, motion expressed in the User Coordinate System

        Args:
            cm (int): control mode
            tx (float): position on the X-axis [mm]
            ty (float): position on the Y-axis [mm]
            tz (float): position on the Z-axis [mm]
            rx (float): rotation around the X-axis [deg]
            ry (float): rotation around the Y-axis [deg]
            rz (float): rotation around the Z-axis [deg]

        Returns:
            0 on success, -1 when ignored, -2 on error.

        Raises:
            PMACError: when the arguments do not match up or when there is a time out or when
            there is a socket
            communication error.

        .. note:: When the command was not successful, this method will query the ``POSVALID?``
                  using the checkAbsolutePosition() and print a summary of the error messages
                  to the log file.
        """

        rc = self.pmac.sendCommand(pmac.CMD_MOVE, cm=cm, tx=tx, ty=ty, tz=tz, rx=rx, ry=ry, rz=rz)

        error_code_msg = {
            0: "Command was successful",
            -1: "Command was ignored",
            -2: "Command was invalid, check with POSVALID?",
        }

        if rc < 0:
            msg = f"Move command returned ({rc}: {error_code_msg[rc]})."

            if rc == -2:
                Q29, errors = self.__check_movement(cm, tx, ty, tz, rx, ry, rz)

                msg += "\nError messages returned from POSVALID?:\n"
                for key, value in errors.items():
                    msg += f"  bit {key:<2d}: {value}\n"

            logger.debug(msg)

        return rc

    def __check_movement(self, cm, tx, ty, tz, rx, ry, rz):
        """
        Ask the controller if the movement defined by the arguments is feasible.

        Returns a tuple where the first element is an integer that represents the
        bitfield encoding the errors. The second element is a dictionary with the
        bit numbers that were (on) and the corresponding error description.

        Args:
            tx (float): position on the X-axis [mm]
            ty (float): position on the Y-axis [mm]
            tz (float): position on the Z-axis [mm]
            rx (float): rotation around the X-axis [deg]
            ry (float): rotation around the Y-axis [deg]
            rz (float): rotation around the Z-axis [deg]

        """
        out = self.pmac.sendCommand(
            pmac.CMD_POSVALID_GET, cm=cm, tx=tx, ty=ty, tz=tz, rx=rx, ry=ry, rz=rz
        )
        Q29 = decode_Q29(out[0])
        return out[0], Q29

    def move_absolute(self, tx, ty, tz, rx, ry, rz):
        try:
            rc = self.__move(0, tx, ty, tz, rx, ry, rz)
        except PMACError as exc:
            raise HexapodError("Couldn't execute the moveAbsolute command.") from exc

        return rc

    def check_absolute_movement(self, tx, ty, tz, rx, ry, rz):
        return self.__check_movement(0, tx, ty, tz, rx, ry, rz)

    def move_relative_object(self, tx, ty, tz, rx, ry, rz):
        try:
            rc = self.__move(1, tx, ty, tz, rx, ry, rz)
        except PMACError as exc:
            raise HexapodError("Couldn't execute the relative movement [object] command.") from exc

        return rc

    def check_relative_object_movement(self, tx, ty, tz, rx, ry, rz):
        return self.__check_movement(1, tx, ty, tz, rx, ry, rz)

    def move_relative_user(self, tx, ty, tz, rx, ry, rz):
        try:
            rc = self.__move(2, tx, ty, tz, rx, ry, rz)
        except PMACError as exc:
            raise HexapodError("Couldn't execute the relative movement [user] command.") from exc

        return rc

    def check_relative_user_movement(self, tx, ty, tz, rx, ry, rz):
        return self.__check_movement(2, tx, ty, tz, rx, ry, rz)

    def perform_maintenance(self, axis):
        try:
            rc = self.pmac.sendCommand(pmac.CMD_MAINTENANCE, axis=axis)
        except PMACError as exc:
            raise HexapodError("Couldn't perform maintenance cycle.") from exc

        msg = {0: "Command was successfully executed", -1: "Command was ignored"}  # noqa: F841

        return rc

    def goto_specific_position(self, pos):
        try:
            rc = self.pmac.sendCommand(pmac.CMD_SPECIFICPOS, pos=pos)
        except PMACError as exc:
            raise HexapodError(f"Couldn't goto specific position [pos={pos}].") from exc

        msg = {
            0: "Command was successfully executed",
            -1: "Command was ignored",
            -2: "Invalid movement command",
        }

        logger.info(f"Goto Specific Position [{pos}]: {msg[rc]}")

        if rc < 0:
            try:
                out = self.pmac.getQVars(0, [29], int)
            except PMACError as exc:
                raise HexapodError("Couldn't get a response from the Hexapod controller.") from exc
            Q29 = decode_Q29(out[0])

            msg = "Error messages returned in Q29:\n"
            for key, value in Q29.items():
                msg += f"  {key:2d}: {value}\n"

            logger.debug(msg)

        return rc

    def goto_retracted_position(self):
        try:
            rc = self.pmac.sendCommand(pmac.CMD_SPECIFICPOS, pos=self.SPEC_POS_RETRACTED)
        except PMACError as exc:
            raise HexapodError("Couldn't goto retracted position.") from exc

        msg = {
            0: "Command was successfully executed",
            -1: "Command was ignored",
            -2: "Invalid movement command",
        }

        logger.info(f"Goto Retracted Position [2]: {msg[rc]}")

        if rc < 0:
            try:
                out = self.pmac.getQVars(0, [29], int)
            except PMACError as exc:
                raise HexapodError("Couldn't get a response from the Hexapod controller.") from exc
            Q29 = decode_Q29(out[0])

            msg = "Error messages returned in Q29:\n"
            for key, value in Q29.items():
                msg += f"  {key:2d}: {value}\n"

            logger.debug(msg)

        return rc

    def goto_zero_position(self):
        try:
            rc = self.pmac.sendCommand(pmac.CMD_SPECIFICPOS, pos=self.SPEC_POS_ZERO)
        except PMACError as exc:
            raise HexapodError("Couldn't goto zero position.") from exc

        msg = {
            0: "Command was successfully executed",
            -1: "Command was ignored",
            -2: "Invalid movement command",
        }

        logger.info(f"Goto Zero Position [1]: {msg[rc]}")

        if rc < 0:
            try:
                out = self.pmac.getQVars(0, [29], int)
            except PMACError as exc:
                raise HexapodError("Couldn't get a response from the Hexapod controller.") from exc
            Q29 = decode_Q29(out[0])

            msg = "Error messages returned in Q29:\n"
            for key, value in Q29.items():
                msg += f"  {key:2d}: {value}\n"

            logger.debug(msg)

        return rc

    def get_buffer(self):
        return_string = self.pmac.getBuffer()
        return return_string

    def clear_error(self):
        try:
            rc = self.pmac.sendCommand(pmac.CMD_CLEARERROR)
        except PMACError as exc:
            raise HexapodError("Couldn't clear errors in the controller software.") from exc

        return rc

    def jog(self, axis: int, inc: float) -> int:
        if not (1 <= axis <= 6):
            logger.error(f"The axis argument must be 1 <= axis <= 6, given {axis}.")
            raise HexapodError("Illegal Argument Value: axis is {axis}, should be 1 <= axis <= 6.")

        try:
            rc = self.pmac.sendCommand(pmac.CMD_JOG, axis=axis, inc=inc)
        except PMACError as exc:
            raise HexapodError(
                f"Couldn't execute the jog command for axis={axis} with inc={inc} [mm]."
            ) from exc

        msg = {0: "Command was successfully executed", -1: "Command was ignored"}

        logger.info(f"JOG on axis [{axis}] of {inc} mm: {msg[rc]}")

        return rc

    def configure_coordinates_systems(
        self, tx_u, ty_u, tz_u, rx_u, ry_u, rz_u, tx_o, ty_o, tz_o, rx_o, ry_o, rz_o
    ):
        try:

            rc = self.pmac.sendCommand(
                pmac.CMD_CFG_CS,
                tx_u=round(tx_u, NUM_OF_DECIMALS),
                ty_u=round(ty_u, NUM_OF_DECIMALS),
                tz_u=round(tz_u, NUM_OF_DECIMALS),
                rx_u=round(rx_u, NUM_OF_DECIMALS),
                ry_u=round(ry_u, NUM_OF_DECIMALS),
                rz_u=round(rz_u, NUM_OF_DECIMALS),
                tx_o=round(tx_o, NUM_OF_DECIMALS),
                ty_o=round(ty_o, NUM_OF_DECIMALS),
                tz_o=round(tz_o, NUM_OF_DECIMALS),
                rx_o=round(rx_o, NUM_OF_DECIMALS),
                ry_o=round(ry_o, NUM_OF_DECIMALS),
                rz_o=round(rz_o, NUM_OF_DECIMALS),
            )
        except PMACError as exc:
            raise HexapodError(
                "Couldn't configure coordinate systems on the hexapod controller."
            ) from exc

        return rc

    def get_coordinates_systems(self):
        try:
            out = self.pmac.sendCommand(pmac.CMD_CFG_CS_GET)
        except PMACError as exc:
            raise HexapodError(
                "Couldn't get the coordinate systems information from the hexapod controller."
            ) from exc

        return out

    def get_debug_info(self):
        try:
            out = self.pmac.sendCommand(pmac.CMD_STATE_DEBUG_GET)
        except PMACError as exc:
            raise HexapodError(
                "Couldn't get the debugging information from the hexapod controller."
            ) from exc

        return out

    def set_speed(self, vt, vr):
        try:
            rc = self.pmac.sendCommand(pmac.CMD_CFG_SPEED, vt=vt, vr=vr)
        except PMACError as exc:
            raise HexapodError(
                "Couldn't set the speed for translation [{vt} mm/s] or rotation [{vr} deg/s]."
            ) from exc

        return rc

    def get_speed(self):
        try:
            out = self.pmac.sendCommand(pmac.CMD_CFG_SPEED_GET)
        except PMACError as exc:
            raise HexapodError(
                "Couldn't get the speed settings from the hexapod controller."
            ) from exc

        return out

    def get_general_state(self):
        try:
            out = self.pmac.getQVars(36, [0], int)
        except PMACError as pmac_exc:
            logger.error(f"PMAC Exception: {pmac_exc}", exc_info=True)
            return None

        return out[0], pmac.decode_Q36(out[0])

    def get_actuator_state(self):
        try:
            out = self.pmac.getQVars(30, [0, 1, 2, 3, 4, 5], int)
        except PMACError as pmac_exc:
            logger.error(f"PMAC Exception: {pmac_exc}", exc_info=True)
            return None

        return [pmac.decode_Q30(value) for value in out]

    def get_user_positions(self):
        try:
            # out = self.pmac.getQVars(53, [0, 1, 2, 3, 4, 5], float)
            out = self.pmac.sendCommand(pmac.CMD_POSUSER_GET)
        except PMACError as pmac_exc:
            logger.error(f"PMAC Exception: {pmac_exc}", exc_info=True)
            return None

        return out

    def get_machine_positions(self):
        try:
            out = self.pmac.getQVars(47, [0, 1, 2, 3, 4, 5], float)
        except PMACError as pmac_exc:
            logger.error(f"PMAC Exception: {pmac_exc}", exc_info=True)
            return None

        return out

    def get_actuator_length(self):
        try:
            out = self.pmac.getQVars(41, [0, 1, 2, 3, 4, 5], float)
        except PMACError as pmac_exc:
            logger.error(f"PMAC Exception: {pmac_exc}", exc_info=True)
            return None

        return out

    def reset(self, wait=True, verbose=False):
        try:
            self.pmac.sendCommand(pmac.CMD_RESETSOFT)
        except PMACError as exc:
            raise HexapodError("Couldn't (soft) reset the hexapod controller.") from exc

        # How do I know when the RESETSOFT has finished and we can send further commands?

        if wait:
            logger.info("Sent a soft reset, this will take about 30 seconds to complete.")
            self.__wait(30, verbose=verbose)

    def __wait(self, duration, verbose=False):
        """
        Wait for a specific duration in seconds.
        """
        _timeout = timedelta(seconds=duration)
        _start = datetime.now()

        _rate = timedelta(seconds=5)  # every _rate seconds print a message
        _count = 0

        logger.info(f"Just waiting {duration} seconds ...")

        while datetime.now() - _start < _timeout:

            if verbose and (datetime.now() - _start > _count * _rate):
                _count += 1
                logger.info(f"waited for {_count * _rate} of {_timeout} seconds, ")
                print(f"waited for {_count * _rate} of {_timeout} seconds, ")

            time.sleep(0.01)


class PunaSimulator(HexapodSimulator, DeviceInterface):
    """
    HexapodSimulator simulates the Symétrie Hexapod PUNA. The class is heavily based on the
    ReferenceFrames in the `egse.coordinates` package.

    The simulator implements the same methods as the HexapodController class which acts on the
    real hardware controller in either simulation mode or with a real Hexapod PUNA connected.

    Therefore, the HexapodSimulator can be used instead of the Hexapod class in test harnesses
    and when the hardware is not available.

    This class simulates all the movements and status of the Hexapod.
    """

    def __init__(self):
        super().__init__()


class PunaProxy(Proxy, PunaInterface):
    """The PunaProxy class is used to connect to the control server and send commands to the
    Hexapod PUNA remotely.

    Args:
        protocol: the transport protocol
        hostname: location of the control server (IP address)
        port: TCP port on which the control server is listening for commands

    """

    def __init__(self, protocol: str, hostname: str, port: int):
        super().__init__(connect_address(protocol, hostname, port))

    @classmethod
    def from_identifier(cls, device_id: str):

        with RegistryClient() as reg:
            service = reg.discover_service(device_id)

            if service:
                protocol = service.get('protocol', 'tcp')
                hostname = service['host']
                port = service['port']

            else:
                raise RuntimeError(f"No service registered as {device_id}")

        logger.info(f"{protocol=}:{hostname=}:{port=}")

        return cls(protocol, hostname, port)



import rich

from egse.hexapod.symetrie.puna import PunaProxy
from egse.registry.client import RegistryClient


def test_puna_cs_with_proxy():

    with RegistryClient() as reg:
        rich.print(reg.list_services())

    puna = PunaProxy()
    puna.connect()



__all__ = [
    "show_processes",
]

from egse.process import ps_egrep


def show_processes():
    """Show the lines from the `ps -ef` command that match processes from this package."""
    return ps_egrep("(joran|puna|zonda)_(ui|cs|sim)")



"""
Registry Service – core service


"""
from __future__ import annotations

import asyncio
import json
import logging
import multiprocessing
import signal
import sys
import textwrap
import time
import uuid
from typing import Any
from typing import Callable

import typer
import zmq
import zmq.asyncio

from egse.registry import DEFAULT_RS_DB_PATH
from egse.registry import DEFAULT_RS_HB_PORT
from egse.registry import DEFAULT_RS_PUB_PORT
from egse.registry import DEFAULT_RS_REQ_PORT
from egse.registry import MessageType
from egse.registry import logger
from egse.registry.backend import AsyncRegistryBackend
from egse.registry.backend import AsyncSQLiteBackend
from egse.registry.client import AsyncRegistryClient
from egse.system import TyperAsyncCommand
from egse.system import get_logging_level

app = typer.Typer(name="rs_cs", no_args_is_help=True)


class AsyncRegistryServer:
    """
    Asynchronous ZeroMQ-based service registry server.

    This server uses the ZeroMQ async API and asyncio for non-blocking operations.

    Args:
        req_port: Port for the service requests socket [default=4242]
        pub_port: Port for the service notifications socket [default=4243]
        hb_port: Port for receiving heartbeats [default=4244]
        backend: a registry backend, [default=AsyncSQLiteBackend]
        db_path: Path to the SQLite database file [default='service_registry.db']
        cleanup_interval: How often to clean up expired services (seconds) [default=10]
    """

    def __init__(
            self,
            req_port: int = DEFAULT_RS_REQ_PORT,
            pub_port: int = DEFAULT_RS_PUB_PORT,
            hb_port: int = DEFAULT_RS_HB_PORT,
            backend: AsyncRegistryBackend | None = None,
            db_path: str = DEFAULT_RS_DB_PATH,
            cleanup_interval: int = 10
    ):
        self.req_port = req_port
        self.pub_port = pub_port
        self.hb_port = hb_port
        self.db_path = db_path
        self.cleanup_interval = cleanup_interval
        self.logger = logger

        self.context = None
        self.req_socket = None
        self.pub_socket = None
        self.hb_socket = None

        # Initialize the storage backend
        self.backend = backend or AsyncSQLiteBackend(db_path)

        # Running flag and event for clean shutdown
        self._running = False
        self._shutdown_event = asyncio.Event()

        # Tasks
        self._tasks = set()

    async def setup_sockets(self):
        """Set up the communication sockets."""

        # Set ZeroMQ to use asyncio
        self.context = zmq.asyncio.Context()

        # Socket to handle requests and commands
        req_rep_endpoint = f"tcp://*:{self.req_port}"
        self.req_socket = self.context.socket(zmq.ROUTER)
        self.req_socket.bind(req_rep_endpoint)
        self.logger.info(f"Binding request socket to {req_rep_endpoint}")

        # Socket to publish service events
        pub_endpoint = f"tcp://*:{self.pub_port}"
        self.pub_socket = self.context.socket(zmq.PUB)
        self.pub_socket.bind(pub_endpoint)
        self.logger.info(f"Binding publish socket to {pub_endpoint}")

        # Socket to handle heartbeats
        hb_endpoint = f"tcp://*:{self.hb_port}"
        self.hb_socket = self.context.socket(zmq.REP)
        self.hb_socket.bind(hb_endpoint)
        self.logger.info(f"Binding heartbeat socket to {hb_endpoint}")

    async def initialize_backend(self):
        """Initialize the storage backend."""
        await self.backend.initialize()

    async def start(self):
        """Start the registry server."""

        multiprocessing.current_process().name = "rs_cs"

        if self._running:
            return

        await self.setup_sockets()

        await self.initialize_backend()

        self._running = True
        self.logger.info(
            f"Async registry server started on ports {self.req_port} (ROUTER-DEALER), {self.pub_port} (PUB), "
            f"and {self.hb_port} (Heartbeat)"
        )

        # Start the cleanup task
        cleanup_task = asyncio.create_task(self._cleanup_loop())
        self._tasks.add(cleanup_task)
        cleanup_task.add_done_callback(self._tasks.discard)

        # Start the request handler task
        request_task = asyncio.create_task(self._handle_requests())
        self._tasks.add(request_task)
        request_task.add_done_callback(self._tasks.discard)

        # Start the heartbeat handler task
        heartbeats_task = asyncio.create_task(self._handle_heartbeats())
        self._tasks.add(heartbeats_task)
        heartbeats_task.add_done_callback(self._tasks.discard)

        # Wait for shutdown
        await self._shutdown_event.wait()

        # Clean shutdown
        await self._shutdown()

    async def _shutdown(self):
        """Perform clean shutdown."""
        self._running = False
        self.logger.info("Shutting down async registry server...")

        # Cancel all tasks
        for task in self._tasks:
            task.cancel()

        # Wait for tasks to complete (with timeout)
        if self._tasks:
            try:
                await asyncio.wait(self._tasks, timeout=2.0)
            except asyncio.CancelledError:
                pass

        # Close database
        await self.backend.close()

        # Close ZeroMQ sockets
        self.req_socket.close()
        self.pub_socket.close()
        self.hb_socket.close()

        # Close context
        self.context.term()

        self.logger.info("Async registry server shutdown complete")

    def stop(self):
        """Signal the server to stop."""
        self._shutdown_event.set()

    async def _cleanup_loop(self):
        """Background task that periodically cleans up expired services."""
        self.logger.info(f"Started cleanup task with interval {self.cleanup_interval}s")

        try:
            while self._running:
                try:
                    # Clean up expired services
                    expired_ids = await self.backend.clean_expired_services()

                    # Publish de-registration events for expired services
                    for service_id in expired_ids:
                        await self._publish_event('expire', {'service_id': service_id})
                except Exception as exc:
                    self.logger.error(f"Error in cleanup task: {exc}")

                # Sleep for the specified interval
                await asyncio.sleep(self.cleanup_interval)
        except asyncio.CancelledError:
            self.logger.info("Cleanup task cancelled")

    async def _handle_requests(self):
        """Task that handles incoming requests."""
        self.logger.info("Started request handler task")

        try:
            while self._running:
                try:
                    # Wait for a request with timeout to allow checking if still running
                    try:
                        # self.logger.info("Waiting for a request with 1s timeout...")
                        message_parts = await asyncio.wait_for(
                            self.req_socket.recv_multipart(),
                            timeout=1.0
                        )
                    except asyncio.TimeoutError:
                        self.logger.debug("waiting for command request...")
                        continue

                    if len(message_parts) >= 3:
                        client_id = message_parts[0]
                        message_type = MessageType(message_parts[1])
                        message_data = message_parts[2]

                        self.logger.info(f"{client_id = }, {message_type = }, {message_data = }")
                        response = await self._process_request(message_data)

                        await self._send_response(client_id, message_type, response)

                except zmq.ZMQError as exc:
                    self.logger.error(f"ZMQ error: {exc}", exc_info=True)
                except Exception as exc:
                    self.logger.error(f"Error handling request: {exc}", exc_info=True)
        except asyncio.CancelledError:
            self.logger.warning("Request handler task cancelled")

    async def _publish_event(self, event_type: str, data: dict[str, Any]):
        """
        Publish an event to subscribers.

        Args:
            event_type: Type of event (register, deregister, expire, etc.)
            data: Event payload
        """
        event = {
            'type': event_type,
            'timestamp': int(time.time()),
            'data': data
        }

        try:
            # Prefix with event type for subscribers that filter by type
            await self.pub_socket.send_multipart(
                [
                    event_type.encode('utf-8'),
                    json.dumps(event).encode('utf-8')
                ]
            )
            self.logger.debug(f"Published {event_type} event: {data}")
        except Exception as exc:
            self.logger.error(f"Failed to publish event: {exc}")

    async def _process_request(self, msg_data: bytes):
        """
        Process a client request and generate a response.

        Args:
            msg_data: the actual JSON with the request

        """
        try:
            request = json.loads(msg_data.decode())
            self.logger.info(f"Received request: {request}")

        except json.JSONDecodeError as exc:
            self.logger.error(f"Invalid JSON received: {exc}")
            return {'success': False, 'error': 'Invalid JSON format'}

        action = request.get('action')
        if not action:
            return {'success': False, 'error': 'Missing required field: action'}

        handlers: dict[str, Callable] = {
            'register': self._handle_register,
            'deregister': self._handle_deregister,
            'renew': self._handle_renew,
            'info': self._handle_info,
            'get': self._handle_get,
            'list': self._handle_list,
            'discover': self._handle_discover,
            'health': self._handle_health,
            'terminate': self._handle_terminate,
        }

        handler = handlers.get(action)
        if not handler:
            return {'success': False, 'error': f'Unknown action: {action}'}

        return await handler(request)

    async def _send_response(self, client_id: bytes, msg_type: MessageType, response: dict[str, Any]):
        """
        If the client expects a reply, send the response.

        Args:
            client_id: the client identification, part 1 of the multipart message
            msg_type: the type of the message, e.g. if reply is required
            response: a dictionary with the status and response

        """
        if msg_type == MessageType.REQUEST_WITH_REPLY:
            await self.req_socket.send_multipart([client_id, MessageType.RESPONSE.value, json.dumps(response).encode()])

    async def _handle_register(self, request: dict[str, Any]) -> dict[str, Any]:
        """Handle a service registration request."""
        self.logger.info(f"Handle registration request: {request}")

        if 'service_info' not in request:
            return {'success': False, 'error': 'Missing required field: service_info'}

        service_info = request['service_info']

        # Validate required fields
        required_fields = ['name', 'host', 'port']
        for field in required_fields:
            if field not in service_info:
                return {'success': False, 'error': f'Missing required field in service_info: {field}'}

        # Generate ID if not provided
        service_id = service_info.get('id')
        if not service_id:
            service_id = f"{service_info['name']}-{uuid.uuid4()}"
            service_info['id'] = service_id

        # Get TTL
        ttl = request.get('ttl', 30)

        # Register the service
        success = await self.backend.register(service_id, service_info, ttl)

        if success:
            # Publish registration event
            await self._publish_event(
                'register', {
                    'service_id': service_id,
                    'service_info': service_info
                }
            )

            return {
                'success': True,
                'service_id': service_id,
                'message': 'Service registered successfully'
            }

        return {'success': False, 'error': 'Failed to register service'}

    async def _handle_deregister(self, request: dict[str, Any]) -> dict[str, Any]:
        """Handle a service de-registration request."""
        service_id = request.get('service_id')

        self.logger.info(f"Handle de-registration request: {request}")

        if not service_id:
            return {'success': False, 'error': 'Missing required field: service_id'}

        # Get service details before de-registering (for event)
        service_info = await self.backend.get_service(service_id)

        # Deregister the service
        success = await self.backend.deregister(service_id)

        if success:
            # Publish de-registration event
            await self._publish_event(
                'deregister', {
                    'service_id': service_id,
                    'service_info': service_info
                }
            )

            return {
                'success': True,
                'message': 'Service deregistered successfully'
            }

        return {'success': False, 'error': 'Service not found or could not be deregistered'}

    async def _handle_renew(self, request: dict[str, Any]) -> dict[str, Any]:
        """Handle a service heartbeat request."""
        service_id = request.get('service_id')

        self.logger.info(f"Handle renew request: {request}")

        if not service_id:
            return {'success': False, 'error': 'Missing required field: service_id'}

        # Renew the service
        success = await self.backend.renew(service_id)

        if success:
            return {
                'success': True,
                'message': 'Service renewed successfully'
            }

        return {'success': False, 'error': 'Service not found or could not be renewed'}

    async def _handle_heartbeats(self):
        """Task that handles heartbeat messages."""
        self.logger.info("Started heartbeats handler task")

        try:
            while self._running:
                try:
                    # Receive heartbeat (non-blocking with timeout)
                    message_json = await asyncio.wait_for(
                        self.hb_socket.recv_string(),
                        timeout=1.0
                    )

                    # Parse the request
                    request = json.loads(message_json)
                    self.logger.info(f"Received heartbeat request: {request}")

                    response = await self._handle_renew(request)

                    # Send the response
                    await self.hb_socket.send_string(json.dumps(response))

                except asyncio.TimeoutError:
                    self.logger.debug("waiting for heartbeat...")
                    continue

                except Exception as exc:
                    self.logger.error(f"Error handling heartbeat request: {exc}")
                    try:
                        await self.hb_socket.send_string(
                            json.dumps(
                                {
                                    'success': False,
                                    'error': str(exc)
                                }
                            )
                        )
                    except Exception:
                        pass

        except asyncio.CancelledError:
            self.logger.info("Heartbeats handler task cancelled")

    async def _handle_get(self, request: dict[str, Any]) -> dict[str, Any]:
        """Handle a request to get a specific service."""

        service_id = request.get('service_id')

        self.logger.info(f"Handle get request: {request}")

        if not service_id:
            return {'success': False, 'error': 'Missing required field: service_id'}

        # Get the service
        service = await self.backend.get_service(service_id)

        if service:
            return {
                'success': True,
                'service': service
            }

        return {'success': False, 'error': 'Service not found'}

    async def _handle_list(self, request: dict[str, Any]) -> dict[str, Any]:
        """Handle a request to list services."""
        service_type = request.get('service_type')

        self.logger.info(f"Handle list request: {request}")

        # List the services
        services = await self.backend.list_services(service_type)

        return {
            'success': True,
            'services': services,
        }

    async def _handle_discover(self, request: dict[str, Any]) -> dict[str, Any]:
        """Handle a service discovery request."""
        service_type = request.get('service_type')

        self.logger.info(f"Handle discover request for service type: {service_type}")

        if not service_type:
            return {'success': False, 'error': 'Missing required field: service_type'}

        # Discover a service
        service = await self.backend.discover_service(service_type)

        if service:
            return {
                'success': True,
                'service': service
            }

        return {'success': False, 'error': f'No healthy service of type {service_type} found'}

    async def _handle_info(self, request: dict[str, Any]) -> dict[str, Any]:
        """Handle the info request and send information about the registry server."""

        self.logger.info(f"Handle info request: {request}")

        # List the services
        services = await self.backend.list_services()

        return {
            'success': True,
            'status': 'ok',
            'req_port': self.req_port,
            'pub_port': self.pub_port,
            'services': services,
        }

    async def _handle_health(self, request: dict[str, Any]) -> dict[str, Any]:
        """Handle a health check request."""

        self.logger.info(f"Handle health request: {request}")

        return {
            'success': True,
            'status': 'ok',
            'timestamp': int(time.time())
        }

    async def _handle_terminate(self, request: dict[str, Any]) -> dict[str, Any]:
        """Handle a termination request."""

        self.logger.info(f"Handle termination request: {request}")

        self.stop()

        return {
            'success': True,
            'status': 'terminating',
            'timestamp': time.time(),
        }


@app.command(cls=TyperAsyncCommand)
async def start(
        req_port: int = DEFAULT_RS_REQ_PORT,
        pub_port: int = DEFAULT_RS_PUB_PORT,
        hb_port: int = DEFAULT_RS_HB_PORT,
        db_path: str = DEFAULT_RS_DB_PATH,
        cleanup_interval: int = 10,
        log_level: str = "WARNING",
):
    """Run the registry server with signal handling."""

    logging.basicConfig(
        level=get_logging_level(log_level),
        format="[%(asctime)s] %(threadName)-12s %(levelname)-8s %(name)-12s %(lineno)5d:%(module)-20s %(message)s",
    )
    logging.getLogger('aiosqlite').setLevel(logging.INFO)

    server = AsyncRegistryServer(
        req_port=req_port,
        pub_port=pub_port,
        hb_port=hb_port,
        db_path=db_path,
        cleanup_interval=cleanup_interval,
    )

    # Set up signal handlers
    loop = asyncio.get_running_loop()

    for sig in (signal.SIGINT, signal.SIGTERM):
        loop.add_signal_handler(
            sig,
            lambda: asyncio.create_task(handle_signal(server))
        )

    # Start server
    await server.start()


async def handle_signal(server):
    """Handle termination signals."""
    logger.info("Received termination signal")
    server.stop()


@app.command(cls=TyperAsyncCommand)
async def status(
        req_port: int = DEFAULT_RS_REQ_PORT,
        pub_port: int = DEFAULT_RS_PUB_PORT,
        hb_port: int = DEFAULT_RS_HB_PORT,
        host: str = "localhost"
):

    with AsyncRegistryClient(
            registry_req_endpoint=f"tcp://{host}:{req_port}",
            registry_sub_endpoint=f"tcp://{host}:{pub_port}",
            registry_hb_endpoint=f"tcp://{host}:{hb_port}",
            request_timeout=5000  # 5 second timeout
    ) as client:
        response = await client.server_status()

    if response['success']:
        status_report = textwrap.dedent(
            f"""\
            Registry Service:
                Status: {response['status']}
                Requests port: {response['req_port']}
                Notifications port: {response['pub_port']}
                Registrations: {", ".join([f"({x['name']}, {x['health']})" for x in response['services']])}\
            """
        )
    else:
        status_report = "Registry Service: not active"

    print(status_report)


@app.command(cls=TyperAsyncCommand)
async def stop(
        req_port: int = DEFAULT_RS_REQ_PORT,
        pub_port: int = DEFAULT_RS_PUB_PORT,
        hb_port: int = DEFAULT_RS_HB_PORT,
        host: str = "localhost"
):

    with AsyncRegistryClient(
            registry_req_endpoint=f"tcp://{host}:{req_port}",
            registry_sub_endpoint=f"tcp://{host}:{pub_port}",
            registry_hb_endpoint=f"tcp://{host}:{hb_port}",
            request_timeout=5000  # 5 second timeout
    ) as client:
        response = await client.terminate_registry_server()

    if response:
        logger.info("Service registry server terminated.")


if __name__ == "__main__":

    try:
        rc = app()
    except zmq.ZMQError as exc:
        if "Address already in use" in str(exc):
            logger.error(f"The Service Registry server is already running: {exc}")
        else:
            logger.error("Couldn't start service registry server", exc_info=True)
        rc = -1

    sys.exit(rc)



from __future__ import annotations

import asyncio
import json
import threading
import time
import uuid
from contextlib import asynccontextmanager
from typing import Any
from typing import Callable
from typing import Coroutine
from typing import Union

import zmq
import zmq.asyncio

from egse.registry import DEFAULT_RS_HB_PORT
from egse.registry import DEFAULT_RS_PUB_PORT
from egse.registry import DEFAULT_RS_REQ_PORT
from egse.registry import MessageType
from egse.registry import logger
from egse.system import do_every

REQUEST_TIMEOUT = 5000


class RegistryClient:
    """
    Synchronous client for the service registry.
    """
    def __init__(
        self,
        registry_req_endpoint: str = None,
        registry_sub_endpoint: str = None,
        registry_hb_endpoint: str = None,
        request_timeout: int = REQUEST_TIMEOUT,
        client_id: str = "registry-client"
    ):
        """
        Initialize the async registry client.

        Args:
            registry_req_endpoint: ZeroMQ endpoint for REQ-REP socket, defaults to DEFAULT_RS_REQ_PORT on localhost.
            registry_sub_endpoint: ZeroMQ endpoint for SUB socket, defaults to DEFAULT_RS_PUB_PORT on localhost.
            registry_hb_endpoint: ZeroMQ endpoint for Heartbeat socket, defaults to DEFAULT_RS_HB_PORT on localhost.
            request_timeout: Timeout for requests in milliseconds, defaults to 5000.
            client_id: client identification, default='registry-client'
        """
        self.registry_req_endpoint = registry_req_endpoint or f"tcp://localhost:{DEFAULT_RS_REQ_PORT}"
        self.registry_sub_endpoint = registry_sub_endpoint or f"tcp://localhost:{DEFAULT_RS_PUB_PORT}"
        self.registry_hb_endpoint = registry_hb_endpoint or f"tcp://localhost:{DEFAULT_RS_HB_PORT}"

        self.request_timeout = request_timeout
        self.logger = logger

        # Service state
        self._service_id = None
        self._service_info = None
        self._ttl = None

        self._hb_stop_event = None

        self._client_id = f"{client_id}-{uuid.uuid4()}".encode()

        self.context: zmq.Context = zmq.Context.instance()

        self.req_socket: zmq.Socket | None = None
        self.sub_socket: zmq.Socket | None = None
        self.hb_socket: zmq.Socket | None = None

        self.poller = zmq.Poller()

    def __enter__(self):
        self.connect()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.disconnect()

    def connect(self):
        # self.logger.debug("Connecting to service registry...")

        # REQ socket for request-reply pattern
        self.req_socket = self.context.socket(zmq.DEALER)
        self.req_socket.setsockopt(zmq.LINGER, 0)  # Don't wait for unsent messages on close()
        self.req_socket.setsockopt(zmq.IDENTITY, self._client_id)
        self.req_socket.connect(self.registry_req_endpoint)

        # SUB socket for receiving events
        self.sub_socket = self.context.socket(zmq.SUB)
        self.sub_socket.connect(self.registry_sub_endpoint)
        # Default to receiving all events
        self.sub_socket.setsockopt_string(zmq.SUBSCRIBE, "")

        self.poller.register(self.req_socket, zmq.POLLIN)
        self.poller.register(self.sub_socket, zmq.POLLIN)

    def _connect_hb_socket(self):
        self.hb_socket = self.context.socket(zmq.REQ)
        self.hb_socket.setsockopt(zmq.LINGER, 0)  # Don't wait for unsent messages on close()
        self.hb_socket.connect(self.registry_hb_endpoint)

    def disconnect(self):
        # self.logger.debug("Disconnecting from service registry...")

        if self.req_socket:
            self.poller.unregister(self.req_socket)
            self.req_socket.close(linger=0)
        self.req_socket = None

        if self.sub_socket:
            self.poller.unregister(self.sub_socket)
            self.sub_socket.close(linger=0)
        self.sub_socket = None

    def _disconnect_hb_socket(self):
        if self.hb_socket:
            self.hb_socket.setsockopt(zmq.LINGER, 0)
            self.hb_socket.close()
        self.hb_socket = None

    def _send_request(self, msg_type: MessageType, request: dict[str, Any]) -> dict[str, Any]:
        """
        Send a request to the registry and get the response.

        Args:
            msg_type: the type of message and reply
            request: The request to send

        Returns:
            The response from the registry.
        """
        try:
            self.logger.debug(f"Sending request: {request}")
            self.req_socket.send_multipart([msg_type.value, json.dumps(request).encode()])

            if msg_type == MessageType.REQUEST_NO_REPLY:
                return {'success': True}

            if self.poller.poll(timeout=self.request_timeout):

                message_parts = self.req_socket.recv_multipart()

                if len(message_parts) >= 2:
                    message_type = MessageType(message_parts[0])
                    message_data = message_parts[1]

                    if message_type == MessageType.RESPONSE:
                        response = json.loads(message_data)
                        self.logger.debug(f"Received response: {response}")
                        return response
                    else:
                        return {
                            'success': False,
                            'error': f'unexpected MessageType received: {message_type.name}, {message_data = }'
                        }
                else:
                    return {
                        'success': False,
                        'error': f'not enough parts received: {len(message_parts)}',
                        'data': message_parts,
                    }
            else:
                self.logger.error(f"Request timed out after {self.request_timeout / 1000:.2f}s")
                return {'success': False, 'error': 'Request timed out'}
        except zmq.ZMQError as exc:
            self.logger.error(f"ZMQ error: {exc}", exc_info=True)
            return {'success': False, 'error': str(exc)}
        except Exception as exc:
            self.logger.error(f"Error sending request: {exc}", exc_info=True)
            return {'success': False, 'error': str(exc)}

    def _send_heartbeat(self, request: dict[str, Any]) -> dict[str, Any]:
        """
        Send a heartbeat to the registry and get the response.

        Args:
            request: The request to send

        Returns:
            The response from the registry.
        """

        try:
            self.logger.debug(f"Sending heartbeat request: {request}")
            self.hb_socket.send_string(json.dumps(request))

            ready_sockets, _, _ = zmq.select([self.hb_socket], [], [], timeout=self.request_timeout)

            if ready_sockets:
                response_json = self.hb_socket.recv_string()
                response = json.loads(response_json)
                self.logger.debug(f"Received response: {response}")
                return response
            else:
                self.logger.error(f"Request timed out after {self.request_timeout / 1000:.2f}s")
                self._disconnect_hb_socket()
                time.sleep(1.0)
                self._connect_hb_socket()
                return {'success': False, 'error': 'Request timed out'}

        except zmq.ZMQError as exc:
            self.logger.error(f"ZMQ error: {exc}", exc_info=True)
            return {'success': False, 'error': str(exc)}
        except Exception as exc:
            self.logger.error(f"Error sending request: {exc}", exc_info=True)
            return {'success': False, 'error': str(exc)}

    def register(
            self,
            name: str,
            host: str,
            port: int,
            service_type: str | None = None,
            metadata: dict[str, Any] | None
            = None,
            ttl: int = 30
            ) -> str | None:
        """
        Register this service with the registry.

        Args:
            name: Service name
            host: Service host/IP
            port: Service port
            service_type: Service type (for discovery)
            metadata: Additional service metadata
            ttl: Time-to-live in seconds

        Returns:
            The service ID if successful, None otherwise
        """
        # Prepare service info
        service_info = {
            'name': name,
            'host': host,
            'port': port
        }

        # Add optional fields
        if service_type:
            service_info['type'] = service_type

        if metadata:
            service_info['metadata'] = metadata

        # Prepare tags for easier discovery
        tags = []
        if service_type:
            tags.append(service_type)
        service_info['tags'] = tags

        # Send registration request
        request = {
            'action': 'register',
            'service_info': service_info,
            'ttl': ttl
        }

        response = self._send_request(MessageType.REQUEST_WITH_REPLY, request)

        if response.get('success'):
            # Store service information for later use
            self._service_id = response.get('service_id')
            self._service_info = service_info
            self._ttl = ttl

            self.logger.info(f"Service registered with ID: {self._service_id}")
            return self._service_id
        else:
            self.logger.error(f"Failed to register service: {response.get('error')}")
            return None

    def deregister(self, service_id: str = None) -> bool:
        """
        Deregister this service from the registry.

        Returns:
            True if successful, False otherwise.
        """

        service_id = service_id or self._service_id

        if not service_id:
            self.logger.warning("Cannot deregister: no service is registered")
            return False

        request = {
            'action': 'deregister',
            'service_id': service_id
        }

        response = self._send_request(MessageType.REQUEST_WITH_REPLY, request)

        if response.get('success'):
            self.logger.info(f"Service deregistered: {service_id}")
            self._service_id = None
            self._service_info = None
            self._ttl = None
            return True
        else:
            self.logger.error(f"Failed to deregister service: {response.get('error')}")
            return False

    def discover_service(self, service_type: str, use_cache: bool = False) -> dict[str, Any] | None:
        """
        Discover a service of the specified type. The service is guaranteed to be healthy at the time of discovery.

        The returned information contains:

        - name: the name of the service
        - host: the ip address or hostname of the service
        - port: the port number for requests to the microservice

        Args:
            service_type: Type of service to discover
            use_cache: Whether to use cached service information

        Returns:
            Service information if found, None otherwise
        """
        # Try to use cache first if enabled
        if use_cache:
            self.logger.info("Cache not yet implemented.")

        request = {
            'action': 'discover',
            'service_type': service_type
        }

        response = self._send_request(MessageType.REQUEST_WITH_REPLY, request)

        # self.logger.debug(f"{response = }")

        if response.get('success'):
            service = response.get('service')
            return service
        else:
            self.logger.warning(f"Service discovery failed: {response.get('error')}")
            return None

    def get_service(self, service_id: str, use_cache: bool = True) -> dict[str, Any] | None:
        """
        Get information about a specific service.

        Args:
            service_id: ID of the service to get
            use_cache: Whether to use cached service information

        Returns:
            Service information if found, None otherwise.
        """

        request = {
            'action': 'get',
            'service_id': service_id
        }

        response = self._send_request(MessageType.REQUEST_WITH_REPLY, request)

        if response.get('success'):
            service = response.get('service')
            return service
        else:
            self.logger.warning(f"Get service failed: {response.get('error')}")
            return None

    def list_services(self, service_type: str | None = None) -> list[dict[str, Any]]:
        """
        List all registered services, optionally filtered by type.

        Args:
            service_type: Type of services to list

        Returns:
            List of service information.
        """
        request = {
            'action': 'list',
            'service_type': service_type
        }

        response = self._send_request(MessageType.REQUEST_WITH_REPLY, request)

        if response.get('success'):
            services = response.get('services', [])
            return services
        else:
            self.logger.warning(f"List services failed: {response.get('error')}")
            return []

    def get_endpoint(self, service_type) -> str | None:
        """Returns the endpoint for the given service type."""
        service = self.discover_service(service_type)

        if service:
            protocol = service.get('protocol', 'tcp')
            hostname = service['host']
            port = service['port']

            return f"{protocol}://{hostname}:{port}"
        else:
            return None

    def health_check(self) -> bool:
        """
        Check if the registry server is healthy.

        Returns:
            True if healthy, False otherwise
        """
        request = {
            'action': 'health'
        }

        response = self._send_request(MessageType.REQUEST_WITH_REPLY, request)
        return response.get('success', False)

    def terminate_registry_server(self) -> bool:
        """
        Send a terminate request to the service registry server. Returns True when successful.
        """
        request = {
            'action': 'terminate'
        }
        response = self._send_request(MessageType.REQUEST_WITH_REPLY, request)
        return response.get('success', False)

    def server_status(self) -> dict[str, Any]:
        """
        Requests the status information from the service registry.
        """
        request = {
            'action': 'info',
        }
        response = self._send_request(MessageType.REQUEST_WITH_REPLY, request)
        return response

    def start_heartbeat(self, interval: int | None = None) -> None:

        if not self._service_id:
            self.logger.warning("Cannot start heartbeat: no service is registered")
            return None

        # If interval not specified, use 1/3 of TTL
        if interval is None:
            interval = max(1, self._ttl // 3)

        def send_heartbeat():
            try:
                request = {
                    'action': 'renew',
                    'service_id': self._service_id
                }

                response = self._send_heartbeat(request)

                if not response.get('success'):
                    self.logger.warning(f"Heartbeat failed: {response.get('error')}")

                    # Do a health check
                    if not self.health_check():
                        self.logger.warning("ServiceRegistry not responding.")
                        return

                else:
                    self.logger.debug(response.get("message"))

            except Exception as exc:
                self.logger.error(f"Error sending heartbeat: {exc}")

        self._hb_stop_event = threading.Event()
        timer_thread = threading.Thread(
            target=do_every,
            args=(interval, send_heartbeat),
            kwargs={
                'stop_event': self._hb_stop_event,
                'setup_func': self._connect_hb_socket,
                'teardown_func': self._disconnect_hb_socket
            }
        )
        timer_thread.daemon = True
        timer_thread.start()

    def stop_heartbeat(self) -> None:
        self._hb_stop_event.set()

    def close(self) -> None:
        """Clean up resources."""
        try:
            if hasattr(self, 'req_socket') and self.req_socket:
                self.req_socket.close()

            if hasattr(self, 'sub_socket') and self.sub_socket:
                self.sub_socket.close()

            if hasattr(self, 'context') and self.context:
                self.context.term()
        except Exception as exc:
            self.logger.error(f"Error during cleanup: {exc}")


class AsyncRegistryClient:
    """
    Asynchronous client for interacting with the ZeroMQ-based service registry.

    This class uses asyncio and the ZeroMQ asyncio API for non-blocking operations.
    """

    def __init__(
            self,
            registry_req_endpoint: str = None,
            registry_sub_endpoint: str = None,
            registry_hb_endpoint: str = None,
            request_timeout: int = REQUEST_TIMEOUT,
            client_id: str = "registry-client"
    ):
        """
        Initialize the async registry client.

        Args:
            registry_req_endpoint: ZeroMQ endpoint for REQ-REP socket, defaults to DEFAULT_RS_REQ_PORT on localhost.
            registry_sub_endpoint: ZeroMQ endpoint for SUB socket, defaults to DEFAULT_RS_PUB_PORT on localhost.
            registry_hb_endpoint: ZeroMQ endpoint for heartbeat socket, defaults to DEFAULT_RS_HB_PORT on localhost.
            request_timeout: Timeout for requests in milliseconds, defaults to 5000.
            client_id: client identification, default='registry-client'
        """
        self.registry_req_endpoint = registry_req_endpoint or f"tcp://localhost:{DEFAULT_RS_REQ_PORT}"
        self.registry_sub_endpoint = registry_sub_endpoint or f"tcp://localhost:{DEFAULT_RS_PUB_PORT}"
        self.registry_hb_endpoint = registry_hb_endpoint or f"tcp://localhost:{DEFAULT_RS_HB_PORT}"

        self.request_timeout = request_timeout
        self.logger = logger

        # Service state
        self._service_id = None
        self._service_info = None
        self._ttl = None

        self._client_id = f"{client_id}-{uuid.uuid4()}".encode()

        self.context = zmq.asyncio.Context.instance()

        self._running = False
        self._tasks = set()
        self._heartbeat_task = None
        self._event_listener_task = None

        self._event_handlers = {}

        # Service cache (for discovery)
        self._service_cache = {}
        self._service_cache_lock = None

        self.req_socket: zmq.asyncio.Socket | None = None
        self.sub_socket: zmq.asyncio.Socket | None = None
        self.hb_socket: zmq.asyncio.Socket | None = None

    def __enter__(self):
        self.connect()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.disconnect()

    def connect(self):
        self.logger.debug("Connecting to service registry...")

        # REQ socket for request-reply pattern
        self.req_socket = self.context.socket(zmq.DEALER)
        self.req_socket.setsockopt(zmq.LINGER, 0)  # Don't wait for unsent messages on close()
        self.req_socket.setsockopt(zmq.IDENTITY, self._client_id)
        self.req_socket.connect(self.registry_req_endpoint)

        # SUB socket for receiving events
        self.sub_socket = self.context.socket(zmq.SUB)
        self.sub_socket.connect(self.registry_sub_endpoint)
        # Default to receiving all events
        self.sub_socket.setsockopt_string(zmq.SUBSCRIBE, "")

        self._connect_hb_socket()

    def _connect_hb_socket(self):
        self.hb_socket = self.context.socket(zmq.REQ)
        self.hb_socket.setsockopt(zmq.LINGER, 0)  # Don't wait for unsent messages on close()
        self.hb_socket.connect(self.registry_hb_endpoint)

    def disconnect(self):
        self.logger.debug("Disconnecting from service registry...")

        if self.req_socket:
            self.req_socket.close(linger=0)
        self.req_socket = None

        if self.sub_socket:
            self.sub_socket.close(linger=0)
        self.sub_socket = None

        self._disconnect_hb_socket()

    def _disconnect_hb_socket(self):
        if self.hb_socket:
            self.hb_socket.close(linger=0)
        self.hb_socket = None

    def _get_service_cache_lock(self):
        if self._service_cache_lock is None:
            self._service_cache_lock = asyncio.Lock()
        return self._service_cache_lock

    async def _send_request(
            self, msg_type: MessageType, request: dict[str, Any], timeout: int = None
    ) -> dict[str, Any]:
        """
        Send a request to the registry and get the response.

        Args:
            msg_type: the type of message and reply
            request: The request to send to the service registry server.
            timeout: The number of milliseconds to wait

        Returns:
            The response from the registry as a dictionary.
        """
        self.logger.debug(f"Sending request: {request}")

        timeout = (timeout or self.request_timeout) / 1000
        try:
            await self.req_socket.send_multipart([msg_type.value, json.dumps(request).encode()])

            try:
                message_parts = await asyncio.wait_for(
                    self.req_socket.recv_multipart(), timeout=timeout
                )

                if len(message_parts) >= 2:
                    message_type = MessageType(message_parts[0])
                    message_data = message_parts[1]

                    if message_type == MessageType.RESPONSE:
                        response = json.loads(message_data)
                        self.logger.debug(f"Received response: {response}")
                        return response
                    else:
                        return {
                            'success': False,
                            'error': f'unexpected MessageType received: {message_type.name}, {message_data = }'
                        }
                else:
                    return {
                        'success': False,
                        'error': f'not enough parts received: {len(message_parts)}',
                        'data': message_parts,
                    }
            except asyncio.TimeoutError:
                self.logger.error(f"Request timed out after {timeout:.2f}s")
                return {'success': False, 'error': 'Request timed out'}
        except zmq.ZMQError as exc:
            self.logger.error(f"ZMQ error: {exc}", exc_info=True)
            return {'success': False, 'error': str(exc)}
        except Exception as exc:
            self.logger.error(f"Error sending request: {exc}", exc_info=True)
            return {'success': False, 'error': str(exc)}

    async def _send_heartbeat(self, request: dict[str, Any], timeout: int = None) -> dict[str, Any]:
        """
        Send a heartbreak request to the registry and get the response.

        Args:
            request: The request to send to the service registry server.
            timeout: The number of milliseconds to wait

        Returns:
            The response from the registry as a dictionary.
        """

        timeout = (timeout or self.request_timeout) / 1000
        try:
            # Send the request
            self.logger.debug(f"Sending heartbeat request: {request}")
            await self.hb_socket.send_string(json.dumps(request))

            # Wait for the response with timeout
            try:
                response_json = await asyncio.wait_for(self.hb_socket.recv_string(), timeout=timeout)
                response = json.loads(response_json)
                self.logger.debug(f"Received heartbeat response: {response = }")
                return response
            except asyncio.TimeoutError:
                self.logger.error(f"Heartbeat request timed out after {timeout:.2f}s")
                # Reset the socket to avoid invalid state
                self.hb_socket.close()
                self.hb_socket = self.context.socket(zmq.REQ)
                self.hb_socket.connect(self.registry_req_endpoint)
                return {'success': False, 'error': 'Heartbeat request timed out'}
        except zmq.ZMQError as exc:
            self.logger.error(f"ZMQ error: {exc}", exc_info=True)
            return {'success': False, 'error': str(exc)}
        except Exception as exc:
            self.logger.error(f"Error sending heartbeat request: {exc}", exc_info=True)
            return {'success': False, 'error': str(exc)}

    async def register(
            self,
            name: str,
            host: str,
            port: int,
            service_type: str | None = None,
            metadata: dict[str, Any] | None
            = None,
            ttl: int = 30
            ) -> str | None:
        """
        Register this service with the registry.

        Args:
            name: Service name
            host: Service host/IP
            port: Service port
            service_type: Service type (for discovery)
            metadata: Additional service metadata
            ttl: Time-to-live in seconds

        Returns:
            The service ID if successful, None otherwise
        """
        # Prepare service info
        service_info = {
            'name': name,
            'host': host,
            'port': port
        }

        # Add optional fields
        if service_type:
            service_info['type'] = service_type

        if metadata:
            service_info['metadata'] = metadata

        # Prepare tags for easier discovery
        tags = []
        if service_type:
            tags.append(service_type)
        service_info['tags'] = tags

        # Send registration request
        request = {
            'action': 'register',
            'service_info': service_info,
            'ttl': ttl
        }

        response = await self._send_request(MessageType.REQUEST_WITH_REPLY, request)

        if response.get('success'):
            # Store service information for later use
            self._service_id = response.get('service_id')
            self._service_info = service_info
            self._ttl = ttl

            self.logger.info(f"Service registered with ID: {self._service_id}")
            return self._service_id
        else:
            self.logger.error(f"Failed to register service: {response.get('error')}")
            return None

    async def deregister(self, service_id: str = None) -> bool:
        """
        Deregister this service or the service with service_id from the registry.

        When you register and deregister with the same client instance, you don't have to provide
        the service_id.

        Args:
            service_id: the service identifier that was previously handed out by the ServiceRegistry after
                registration. If not provided, the service_id that was saved when registering is used.

        Returns:
            True if successful, False otherwise
        """

        service_id = service_id or self._service_id

        if not service_id:
            self.logger.warning("Cannot deregister: no service is registered")
            return False

        request = {
            'action': 'deregister',
            'service_id': service_id
        }

        response = await self._send_request(MessageType.REQUEST_WITH_REPLY, request)

        if response.get('success'):
            self.logger.info(f"Service deregistered: {service_id}")
            self._service_id = None
            self._service_info = None
            self._ttl = None
            return True
        else:
            self.logger.error(f"Failed to deregister service: {response.get('error')}")
            return False

    async def start_heartbeat(self, interval: int | None = None) -> asyncio.Task | None:
        """
        Start sending heartbeats to the registry.

        Args:
            interval: Heartbeat interval in seconds (default: 1/3 of TTL)

        Returns:
            The heartbeat task
        """
        if not self._service_id:
            self.logger.warning("Cannot start heartbeat: no service is registered")
            return None

        # Cancel existing heartbeat task if present
        await self.stop_heartbeat()

        # If interval not specified, use 1/3 of TTL
        if interval is None:
            interval = max(1, self._ttl // 3)

        self._running = True

        async def heartbeat_loop():
            try:
                while self._running and self._service_id:
                    try:
                        request = {
                            'action': 'renew',
                            'service_id': self._service_id
                        }

                        response = await self._send_heartbeat(request)

                        if not response.get('success'):
                            self.logger.warning(f"Heartbeat failed: {response.get('error')}")
                        else:
                            self.logger.info(response.get("message"))
                    except Exception as exc:
                        self.logger.error(f"Error in heartbeat loop: {exc}", exc_info=True)

                    # Sleep until next heartbeat
                    await asyncio.sleep(interval)
            except asyncio.CancelledError:
                self.logger.info("Heartbeat task cancelled")

        # Start the heartbeat task
        self._heartbeat_task = task = asyncio.create_task(heartbeat_loop())
        self._tasks.add(task)
        task.add_done_callback(lambda t: self._tasks.discard(t))

        self.logger.info(f"Started heartbeat task with interval {interval}s")
        return task

    async def stop_heartbeat(self) -> None:
        """Stop the running heartbeat task."""
        if self._heartbeat_task is not None:
            self._heartbeat_task.cancel()
            try:
                await self._heartbeat_task
            except asyncio.CancelledError:
                pass
            self._tasks.discard(self._heartbeat_task)
            self._heartbeat_task = None
            self.logger.info("Stopped heartbeat task")
        else:
            self.logger.info("Couldn't stop heartbeat, heartbeat_task is None")

    async def stop_event_listener(self) -> None:
        """Stop the running event listener task."""
        if self._event_listener_task is not None:
            self._event_listener_task.cancel()
            try:
                await self._event_listener_task
            except asyncio.CancelledError:
                pass
            self._tasks.discard(self._event_listener_task)
            self._event_listener_task = None
            self.logger.info("Stopped event listener task")
        else:
            self.logger.info("Couldn't stop event_listener, event_listener_task is None")

    async def stop_all_tasks(self) -> None:
        self._running = False

        # Cancel all tasks
        for task in self._tasks:
            task.cancel()

        # Wait for tasks to complete (with timeout)
        if self._tasks:
            try:
                await asyncio.wait(self._tasks, timeout=2.0)
            except asyncio.CancelledError:
                pass

        self._tasks.clear()
        self.logger.info("Stopped all background tasks")

    def on_event(self, event_type: str, handler: Callable[[dict[str, Any]], Union[None, Coroutine]]) -> None:
        """
        Register a handler for a specific event type.

        Args:
            event_type: Type of event to handle (register, deregister, expire)
            handler: Function or coroutine to call with event data
        """
        if event_type not in self._event_handlers:
            self._event_handlers[event_type] = []

            # Subscribe to this specific event type
            self.sub_socket.setsockopt_string(zmq.SUBSCRIBE, event_type)

        self._event_handlers[event_type].append(handler)
        self.logger.debug(f"Registered handler for {event_type} events")

    async def start_event_listener(self) -> asyncio.Task:
        """
        Start listening for registry events.

        Returns:
            The event listener task
        """
        # Cancel existing event listener task if present
        await self.stop_event_listener()

        self._running = True

        async def subscription_loop():
            try:
                while self._running:
                    try:
                        # Use a timeout to allow for clean shutdown
                        try:
                            message = await asyncio.wait_for(
                                self.sub_socket.recv_multipart(),
                                timeout=1.0
                            )
                        except asyncio.TimeoutError:
                            continue

                        # Parse the message
                        event_type_bytes, event_json_bytes = message
                        event_type = event_type_bytes.decode('utf-8')
                        event = json.loads(event_json_bytes.decode('utf-8'))

                        self.logger.debug(f"Received event: {event_type}")

                        # Update service cache based on events
                        await self._update_cache_from_event(event_type, event)

                        # Call registered handlers
                        handlers = self._event_handlers.get(event_type, [])
                        for handler in handlers:
                            try:
                                # Check if handler is a coroutine function
                                if asyncio.iscoroutinefunction(handler):
                                    await handler(event['data'])
                                else:
                                    handler(event['data'])
                            except Exception as exc:
                                self.logger.error(f"Error in event handler: {exc}")
                    except zmq.ZMQError as exc:
                        self.logger.error(f"ZMQ error in event listener: {exc}")
                    except Exception as exc:
                        self.logger.error(f"Error in event listener: {exc}")
                        await asyncio.sleep(1)  # Prevent tight loop on error
            except asyncio.CancelledError:
                self.logger.info("Event listener task cancelled")

        # Start the subscription task
        self._event_listener_task = task = asyncio.create_task(subscription_loop())
        self._tasks.add(task)
        task.add_done_callback(lambda t: self._tasks.discard(t))

        self.logger.info("Started event listener task")
        return task

    async def _update_cache_from_event(self, event_type: str, event: dict[str, Any]) -> None:
        """
        Update the service cache based on registry events.

        Args:
            event_type: Type of event
            event: Event data
        """
        async with self._get_service_cache_lock():
            data = event.get('data', {})
            service_id = data.get('service_id')

            if not service_id:
                return

            if event_type == 'register':
                service_info = data.get('service_info', {})
                if service_info:
                    self._service_cache[service_id] = service_info
            elif event_type in ('deregister', 'expire'):
                if service_id in self._service_cache:
                    del self._service_cache[service_id]

    async def discover_service(self, service_type: str, use_cache: bool = True) -> dict[str, Any] | None:
        """
        Discover a service of the specified type. The service is guaranteed to be healthy at the time of discovery.

        The returned information contains:

        - name: the name of the service
        - host: the ip address or hostname of the service
        - port: the port number for requests to the microservice

        Args:
            service_type: Type of service to discover
            use_cache: Whether to use cached service information

        Returns:
            Service information if found, None otherwise
        """
        # Try to use cache first if enabled
        if use_cache:
            async with self._get_service_cache_lock():
                # Find services of the specified type
                matching_services = []
                for service_id, service_info in self._service_cache.items():
                    if (service_info.get('type') == service_type or
                            service_type in service_info.get('tags', [])):
                        matching_services.append(service_info)

                if matching_services:
                    # Simple load balancing - random selection
                    import random
                    return random.choice(matching_services)

        # If not found in cache or cache disabled, ask the registry
        request = {
            'action': 'discover',
            'service_type': service_type
        }

        response = await self._send_request(MessageType.REQUEST_WITH_REPLY, request)

        self.logger.debug(f"{response = }")

        if response.get('success'):
            service = response.get('service')

            # Update cache
            if service and 'id' in service:
                async with self._get_service_cache_lock():
                    self._service_cache[service['id']] = service

            return service
        else:
            self.logger.warning(f"Service discovery failed: {response.get('error')}")
            return None

    async def get_service(self, service_id: str, use_cache: bool = True) -> dict[str, Any] | None:
        """
        Get information about a specific service.

        Args:
            service_id: ID of the service to get
            use_cache: Whether to use cached service information

        Returns:
            Service information if found, None otherwise
        """
        # Try to use cache first if enabled
        if use_cache:
            async with self._get_service_cache_lock():
                if service_id in self._service_cache:
                    return self._service_cache[service_id]

        # If not found in cache or cache disabled, ask the registry
        request = {
            'action': 'get',
            'service_id': service_id
        }

        response = await self._send_request(MessageType.REQUEST_WITH_REPLY, request)

        if response.get('success'):
            service = response.get('service')

            # Update cache
            if service:
                async with self._get_service_cache_lock():
                    self._service_cache[service_id] = service

            return service
        else:
            self.logger.warning(f"Get service failed: {response.get('error')}")
            return None

    async def list_services(self, service_type: str | None = None) -> list[dict[str, Any]]:
        """
        List all registered services, optionally filtered by type.

        Args:
            service_type: Type of services to list

        Returns:
            List of service information
        """
        request = {
            'action': 'list',
            'service_type': service_type
        }

        response = await self._send_request(MessageType.REQUEST_WITH_REPLY, request)

        if response.get('success'):
            services = response.get('services', [])

            # Update cache
            async with self._get_service_cache_lock():
                for service in services:
                    if 'id' in service:
                        self._service_cache[service['id']] = service

            return services
        else:
            self.logger.warning(f"List services failed: {response.get('error')}")
            return []

    async def get_endpoint(self, service_type) -> str | None:
        """Returns the endpoint for the given service type."""
        service = await self.discover_service(service_type)

        if service:
            protocol = service.get('protocol', 'tcp')
            hostname = service['host']
            port = service['port']

            return f"{protocol}://{hostname}:{port}"
        else:
            return None

    async def health_check(self) -> bool:
        """
        Check if the registry server is healthy.

        Returns:
            True if healthy, False otherwise
        """
        request = {
            'action': 'health'
        }

        response = await self._send_request(MessageType.REQUEST_WITH_REPLY, request)
        return response.get('success', False)

    async def terminate_registry_server(self) -> bool:
        """
        Send a terminate request to the service registry server. Returns True when successful.
        """
        request = {
            'action': 'terminate'
        }
        response = await self._send_request(MessageType.REQUEST_WITH_REPLY, request)
        return response.get('success', False)

    async def server_status(self) -> dict[str, Any]:
        """
        Requests the status information from the service registry.
        """
        request = {
            'action': 'info',
        }
        response = await self._send_request(MessageType.REQUEST_WITH_REPLY, request, timeout=500)
        return response

    async def close(self) -> None:
        """Clean up resources."""
        await self.stop_heartbeat()  # This stops all tasks

        try:
            if hasattr(self, 'req_socket') and self.req_socket:
                self.req_socket.close()

            if hasattr(self, 'sub_socket') and self.sub_socket:
                self.sub_socket.close()

            # We can not terminate the context, because we use a global instance, i.e. a singleton context.
            # When we try to terminate it, even after checking if it was closed,
            if hasattr(self, 'context') and self.context:
                self.logger.info(f"{self.context = !r}")
                self.logger.info(f"{self.context._sockets = !r}")
                if not self.context.closed:
                    self.context.term()
        except Exception as exc:
            self.logger.error(f"Error during cleanup: {exc}")

    @asynccontextmanager
    async def register_context(self, *args, **kwargs):
        """
        Async context manager for service registration.

        Example:
            async with client.register_context("my-service", "localhost", 8080):
                # Service is registered
                await app.start()
            # Service is automatically deregistered
        """
        service_id = await self.register(*args, **kwargs)

        if not service_id:
            raise RuntimeError("Failed to register service")

        # Start heartbeat and event listener
        await self.start_heartbeat()
        await self.start_event_listener()

        try:
            yield service_id
        finally:
            # Clean up
            await self.stop_event_listener()
            await self.stop_heartbeat()
            await self.deregister()
            await self.close()


def is_service_registered(service_type: str):
    """Convenience function to check if a service is registered."""
    with RegistryClient() as reg:
        response = reg.discover_service(service_type)

    return False if response is None else True
